package pidgio

  1. Overview
  2. Docs

Module Eff.TraversableSource

Traversable interface on top of Req.S1 for S1

Parameters

module T : Primavera.Sig.Req.T1 with type 'a applicative = 'a output

Signature

Sourceval traverse : ('a -> ('b, 'f) t) -> 'a T.foldable -> ('b T.foldable, 'f) t

Map each element of a structure to an action, evaluate these actions from left to right, and collect the results. *

Sourceval sequence : ('a, 'handler) t T.foldable -> ('a T.foldable, 'handler) t

Evaluate each action in the structure from left to right, and collect the results